# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Basic device abstraction for PILS devices."""

import time

from zapf import ApiError, SpecError
from zapf.spec import FIRST_FLOAT_PARAM, ParamCMDs, ParamControl, Parameters, \
    PLCStatus, SpecialFunctions, StatusStruct, FLOAT64_MAX, FLOAT32_MAX
from zapf.util import UncasedMap


START16 = PLCStatus.START << 12
START32 = PLCStatus.START << 28


class Device:
    @classmethod
    def class_for(cls, typecode):
        return PLC_CLASSES.get(typecode)

    valuetype = None

    def __init__(self, name, addr, typecode, params, funcs, io, log):
        self.name = name
        self.addr = addr
        self.target_addr = None
        self.status_addr = None
        self.param_addr = None
        self.device_kind = typecode >> 8
        self.total_size = 2 * (typecode & 0xff)
        self.value_size = None
        self.status_size = None
        self.params = params
        self.funcs = funcs
        self.log = log
        self.io = io
        self.init()
        self.io.register_cache_range(self.addr, self.total_size)

    def read_status(self):
        if self.status_addr is None:
            return StatusStruct(PLCStatus.IDLE << 12)
        if self.status_size == 2:
            return StatusStruct(self.io.read_u16(self.status_addr))
        return StatusStruct(self.io.read_u32(self.status_addr) >> 16)

    def read_extended_status(self):
        if self.status_addr is None:
            return (StatusStruct(PLCStatus.IDLE << 12), 0, 0)
        if self.status_size == 2:
            return (StatusStruct(self.io.read_u16(self.status_addr)), 0, 0)
        elif self.status_size == 4:
            value = self.io.read_u32(self.status_addr)
            return (StatusStruct(value >> 16), value & 0xFFFF, 0)
        elif self.status_size == 8:  # actually 6, but this is easier
            value = self.io.read_u64(self.status_addr)
            return (StatusStruct(value >> 48), (value >> 32) & 0xFFFF,
                    (value >> 16) & 0xFFFF)
        raise SpecError('invalid status_size')

    def change_status(self, initial_states=(), final_state=0):
        if self.status_addr is None:
            return False
        status = self.read_status()
        if status.STATE in initial_states:
            status.STATE = final_state
            if self.status_size == 2:
                self.io.write_u16(self.status_addr, int(status))
            else:
                self.io.write_u32(self.status_addr, int(status) << 16)
            return True
        return False

    # to implement:

    def init(self):
        raise NotImplementedError

    def get_limits(self):
        raise NotImplementedError

    def read_value(self):
        raise NotImplementedError

    def read_target(self):
        raise ApiError('reading target of a read-only device')

    def change_target(self, value):
        raise ApiError('writing target of a read-only device')

    def list_params(self):
        return []

    def get_param(self, name):
        raise ApiError('reading parameter of a device without params')

    def set_param(self, name, value):
        raise ApiError('writing parameter of a device without params')

    def list_funcs(self):
        return []

    def exec_func(self, name, value=None):
        raise ApiError('executing function of a device without functions')


class DiscreteDevice(Device):
    def init(self):
        pass

    def get_limits(self):
        if self.value_size == 2:
            return (-2**15, 2**15 - 1)
        elif self.value_size == 4:
            return (-2**31, 2**31 - 1)
        elif self.value_size == 8:
            return (-2**63, 2**63 - 1)
        raise SpecError('invalid value_size')

    def read_value(self):
        return self.io.read_fmt(self.addr, self.value_fmt)[0]

    def read_target(self):
        if self.target_addr is None:
            raise ApiError('reading target of a read-only device')
        return self.io.read_fmt(self.target_addr, self.value_fmt)[0]

    def change_target(self, value):
        if self.target_addr is None:
            raise ApiError('writing target of a read-only device')
        if self.value_size == 2:
            self.io.write_fmt(self.target_addr, self.value_fmt + 'H',
                              value, START16)
        else:
            self.io.write_fmt(self.target_addr, self.value_fmt + 'I',
                              value, START32)


INT_FMT = {2: 'h', 4: 'i', 8: 'q'}


class SimpleDiscreteIn(DiscreteDevice):
    def init(self):
        self.value_size = self.total_size
        self.value_fmt = INT_FMT[self.value_size]


class SimpleDiscreteOut(DiscreteDevice):
    def init(self):
        self.value_size = self.total_size // 2
        self.value_fmt = INT_FMT[self.value_size]
        self.target_addr = self.addr + self.value_size


class DiscreteIn(DiscreteDevice):
    def init(self):
        self.value_size = self.status_size = self.total_size // 2
        self.value_fmt = INT_FMT[self.value_size]
        self.status_addr = self.addr + self.value_size


class DiscreteOut(DiscreteDevice):
    def init(self):
        self.value_size = self.status_size = self.total_size // 3
        self.value_fmt = INT_FMT[self.value_size]
        self.target_addr = self.addr + self.value_size
        self.status_addr = self.target_addr + self.value_size


class Keyword(DiscreteDevice):
    def init(self):
        self.value_size = self.total_size
        self.value_fmt = INT_FMT[self.value_size].upper()
        self.target_addr = self.addr

    def get_limits(self):
        if self.value_size == 2:
            return (0, 2**16 - 1)
        elif self.value_size == 4:
            return (0, 2**32 - 1)
        elif self.value_size == 8:
            return (0, 2**64 - 1)
        raise SpecError('invalid value_size')


class StatusWord(Keyword):
    def init(self):
        Keyword.init(self)
        self.status_addr = self.addr
        self.status_size = self.value_size


class AnalogDevice(Device):
    def init(self):
        pass

    def get_limits(self):
        if self.value_size == 4:
            return (-FLOAT32_MAX, FLOAT32_MAX)
        elif self.value_size == 8:
            return (-FLOAT64_MAX, FLOAT64_MAX)
        raise SpecError('invalid value_size')

    def read_value(self):
        if self.value_size == 4:
            return self.io.read_f32(self.addr)
        elif self.value_size == 8:
            return self.io.read_f64(self.addr)
        raise SpecError('invalid value_size')

    def read_target(self):
        if self.target_addr is None:
            raise ApiError('reading target of a read-only device')
        if self.value_size == 4:
            return self.io.read_f32(self.target_addr)
        elif self.value_size == 8:
            return self.io.read_f64(self.target_addr)
        raise SpecError('invalid value_size')

    def change_target(self, value):
        if self.target_addr is None:
            raise ApiError('writing target of a read-only device')
        if self.value_size == 4:
            if self.status_size == 2:
                self.io.write_f32_u16(self.target_addr, value, START16)
            else:
                self.io.write_f32_u32(self.target_addr, value, START32)
        elif self.value_size == 8:
            self.io.write_fmt(self.target_addr, 'dI', value, START32)


class SimpleAnalogIn(AnalogDevice):
    def init(self):
        self.value_size = self.total_size


class SimpleAnalogOut(AnalogDevice):
    def init(self):
        self.value_size = self.total_size // 2
        self.target_addr = self.addr + self.value_size


class AnalogIn(AnalogDevice):
    def init(self):
        if self.total_size == 6:
            # 32-bit with 16 bit status
            self.value_size = 4
            self.status_size = 2
        else:
            self.value_size = self.status_size = self.total_size // 2
        self.status_addr = self.addr + self.value_size


class AnalogOut(AnalogDevice):
    def init(self):
        if self.total_size == 10:
            # 32-bit with 16 bit status
            self.value_size = 4
            self.status_size = 2
        else:
            self.value_size = self.status_size = self.total_size // 3
        self.target_addr = self.addr + self.value_size
        self.status_addr = self.target_addr + self.value_size


class RealValue(AnalogDevice):
    def init(self):
        self.value_size = self.total_size
        self.target_addr = self.addr


class FlatParams:
    def init(self):
        if self.number != len(self.params):
            raise SpecError('mismatch between parameter count between '
                            'typecode and parameter indices from indexer '
                            f'({self.number}/{len(self.params)})')
        # sort params from indexer by index
        pars_indices = sorted((Parameters[par], par) for par in self.params)
        self.param_map = UncasedMap(*(
            (par[1], (self.param_addr + i * self.value_size, par[0]))
            for (i, par) in enumerate(pars_indices)
        ))

    def list_params(self):
        return self.param_map.init_keys()

    def get_param(self, name):
        (addr, idx) = self.param_map.get(name, (None, None))
        if addr:
            if idx < FIRST_FLOAT_PARAM:
                if self.value_size == 4:
                    return ParamCMDs.DONE, self.io.read_u32(addr)
                else:
                    return ParamCMDs.DONE, self.io.read_u64(addr)
            else:
                if self.value_size == 4:
                    return ParamCMDs.DONE, self.io.read_f32(addr)
                else:
                    return ParamCMDs.DONE, self.io.read_f64(addr)
        return ParamCMDs.ERR_NO_IDX, None

    def set_param(self, name, value):
        (addr, idx) = self.param_map.get(name, (None, None))
        if addr:
            if idx < FIRST_FLOAT_PARAM:
                if self.value_size == 4:
                    self.io.write_u32(addr, int(value))
                    return ParamCMDs.DONE, self.io.read_u32(addr)
                else:
                    self.io.write_u64(addr, int(value))
                    return ParamCMDs.DONE, self.io.read_u64(addr)
            else:
                if self.value_size == 4:
                    self.io.write_f32(addr, value)
                    return ParamCMDs.DONE, self.io.read_f32(addr)
                else:
                    self.io.write_f64(addr, value)
                    return ParamCMDs.DONE, self.io.read_f64(addr)
        return ParamCMDs.ERR_NO_IDX, None


class FlatIn(FlatParams, AnalogDevice):
    def init(self):
        self.number = (self.device_kind & 0xf) + 1
        self.value_size = self.total_size // (self.number + 2)
        self.status_size = 2 if self.value_size == 4 else 8
        self.status_addr = self.addr + self.value_size
        self.param_addr = self.status_addr + self.status_size
        FlatParams.init(self)


class FlatOut(FlatParams, AnalogDevice):
    def init(self):
        self.number = (self.device_kind & 0xf) + 1
        self.value_size = self.total_size // (self.number + 3)
        self.status_size = 2 if self.value_size == 4 else 8
        self.target_addr = self.addr + self.value_size
        self.status_addr = self.target_addr + self.value_size
        self.param_addr = self.status_addr + self.status_size
        FlatParams.init(self)


class ParamInterface:
    param_timeout = 1

    def init(self):
        self.pctrl_addr = self.status_addr + (2 if self.value_size == 4 else 6)
        self.param_addr = self.pctrl_addr + 2
        self.param_map = UncasedMap(*((p, Parameters[p])
                                      for p in self.params))
        self.func_map = UncasedMap(*((p, SpecialFunctions[p])
                                     for p in self.funcs))
        self.sm = ParamControl()

    def _update_sm(self):
        self.sm(self.io.read_u16(self.pctrl_addr))

    def _wait_sm_available(self):
        self._update_sm()
        timesout = time.time() + self.param_timeout
        while not self.sm.available:
            self._update_sm()
            if time.time() > timesout:
                return False
        return True

    def _set_param_value(self, idx, value):
        if idx < FIRST_FLOAT_PARAM:
            if self.value_size == 4:
                self.io.write_u32(self.param_addr, int(value))
            else:
                self.io.write_u64(self.param_addr, int(value))
        else:
            if self.value_size == 4:
                self.io.write_f32(self.param_addr, value)
            else:
                self.io.write_f64(self.param_addr, value)

    def _set_param_value_and_cmd(self, cmd, sub, idx, value):
        self.sm.CMD = cmd
        self.sm.SUB = sub
        self.sm.IDX = idx
        if idx < FIRST_FLOAT_PARAM:
            if self.value_size == 4:
                self.io.write_fmt(self.pctrl_addr, 'HI', int(self.sm), int(value))
            else:
                self.io.write_fmt(self.pctrl_addr, 'HQ', int(self.sm), int(value))
        else:
            if self.value_size == 4:
                self.io.write_u16_f32(self.pctrl_addr, int(self.sm), value)
            else:
                self.io.write_fmt(self.pctrl_addr, 'Hd', int(self.sm), value)

    def _set_param_cmd(self, cmd, sub, idx):
        self.sm.CMD = cmd
        self.sm.SUB = sub
        self.sm.IDX = idx
        self.io.write_u16([self.pctrl_addr, int(self.sm)])

    def _get_param_value(self, idx):
        if idx < FIRST_FLOAT_PARAM:
            if self.value_size == 4:
                return self.io.read_u32(self.param_addr)
            else:
                return self.io.read_u64(self.param_addr)
        else:
            if self.value_size == 4:
                return self.io.read_f32(self.param_addr)
            else:
                return self.io.read_f64(self.param_addr)

    def list_params(self):
        return self.param_map.init_keys()

    def get_param(self, name):
        idx = self.param_map.get(name)
        if idx:
            if self._wait_sm_available():
                self._set_param_cmd(ParamCMDs.DO_READ, self.subdev, idx)
                # now wait until setting parameter is finished and
                # return read-back-value
                if self._wait_sm_available():
                    return self.sm.CMD, self._get_param_value(idx)
            return ParamCMDs.ERR_RETRY, None
        return ParamCMDs.ERR_NO_IDX, None

    def set_param(self, name, value):
        idx = self.param_map.get(name)
        if idx:
            if self._wait_sm_available():
                self._set_param_value_and_cmd(ParamCMDs.DO_WRITE,
                                              self.subdev, idx, value)
                # now wait until setting parameter is finished and
                # return read-back-value
                if self._wait_sm_available():
                    return self.sm.CMD, self._get_param_value(idx)
            return ParamCMDs.ERR_RETRY, None
        return ParamCMDs.ERR_NO_IDX, None

    def list_funcs(self):
        return self.func_map.init_keys()

    def exec_func(self, name, value=None):
        idx = self.func_map.get(name)
        if idx:
            self._update_sm()
            if self.sm.CMD == ParamCMDs.BUSY and \
               self.sm.SUB == self.subdev and \
               self.sm.IDX == idx:
                self._set_param_value(idx, value)
                if self._wait_sm_available():
                    return self.sm.CMD, self._get_param_value(idx)
                return self.sm.CMD, self._get_param_value(idx)  # still BUSY
            if self._wait_sm_available():
                self._set_param_value_and_cmd(ParamCMDs.DO_WRITE,
                                              self.subdev, idx, value)
                # now wait until setting parameter is finished and return
                # read-back-value
                if self._wait_sm_available():
                    return self.sm.CMD, self._get_param_value(idx)
                return self.sm.CMD, self._get_param_value(idx)  # still BUSY
            return ParamCMDs.ERR_RETRY, None
        return ParamCMDs.ERR_NO_IDX, None


class ParamIn(ParamInterface, AnalogDevice):
    def init(self):
        self.value_size = self.total_size // 3
        self.status_size = 2 if self.value_size == 4 else 8
        self.status_addr = self.addr + self.value_size
        ParamInterface.init(self)


class ParamOut(ParamInterface, AnalogDevice):
    def init(self):
        self.value_size = self.total_size // 4
        self.status_size = 2 if self.value_size == 4 else 8
        self.target_addr = self.addr + self.value_size
        self.status_addr = self.target_addr + self.value_size
        ParamInterface.init(self)


class VectorDevice(Device):
    def init(self):
        pass

    def get_limits(self):
        if self.value_size == 4:
            return (-FLOAT32_MAX, FLOAT32_MAX)
        elif self.value_size == 8:
            return (-FLOAT64_MAX, FLOAT64_MAX)
        raise SpecError('invalid value_size')

    def read_value(self):
        if self.value_size == 4:
            return self.io.read_f32s(self.addr, self.number)
        elif self.value_size == 8:
            return self.io.read_f64s(self.addr, self.number)
        raise SpecError('invalid value_size')


class VectorIn(ParamInterface, VectorDevice):
    def init(self):
        self.number = (self.device_kind & 0xf) + 1
        self.value_size = self.total_size // (self.number + 2)
        self.status_size = 2 if self.value_size == 4 else 8
        self.status_addr = self.addr + self.value_size*self.number
        ParamInterface.init(self)


class VectorOut(ParamInterface, VectorDevice):
    def init(self):
        self.number = (self.device_kind & 0xf) + 1
        self.value_size = self.total_size // (2*self.number + 2)
        self.status_size = 2 if self.value_size == 4 else 8
        self.target_addr = self.addr + self.value_size*self.number
        self.status_addr = self.target_addr + self.value_size*self.number
        ParamInterface.init(self)

    def read_target(self):
        if self.value_size == 4:
            return self.io.read_f32s(self.target_addr, self.number)
        elif self.value_size == 8:
            return self.io.read_f64s(self.target_addr, self.number)
        raise SpecError('invalid value_size')

    def change_target(self, value):
        if self.value_size == 4:
            self.io.write_f32s_u16(self.target_addr, value, START16)
        elif self.value_size == 8:
            self.io.write_f64s_u32(self.target_addr, value, START32)


PLC_CLASSES = {
    0x1201: SimpleDiscreteIn,   # 16 bit
    0x1202: SimpleDiscreteIn,   # 32 bit
    0x1204: SimpleDiscreteIn,   # 64 bit
    0x1302: SimpleAnalogIn,     # 32 bit
    0x1304: SimpleAnalogIn,     # 64 bit
    0x1401: Keyword,            # 16 bit
    0x1402: Keyword,            # 32 bit
    0x1404: Keyword,            # 64 bit
    0x1502: RealValue,          # 32 bit
    0x1504: RealValue,          # 64 bit
    0x1602: SimpleDiscreteOut,  # 16 bit
    0x1604: SimpleDiscreteOut,  # 32 bit
    0x1608: SimpleDiscreteOut,  # 64 bit
    0x1704: SimpleAnalogOut,    # 32 bit
    0x1708: SimpleAnalogOut,    # 64 bit
    0x1801: StatusWord,         # simple
    0x1802: StatusWord,         # extended
    0x1a02: DiscreteIn,         # 16 bit
    0x1a04: DiscreteIn,         # 32 bit
    0x1a08: DiscreteIn,         # 64 bit
    0x1b03: AnalogIn,           # 32 bit with 16 bit status
    0x1b04: AnalogIn,           # 32 bit
    0x1b08: AnalogIn,           # 64 bit
    0x1e03: DiscreteOut,        # 16 bit
    0x1e06: DiscreteOut,        # 32 bit
    0x1e0c: DiscreteOut,        # 64 bit
    0x1f05: AnalogOut,          # 32 bit with 16 bit status
    0x1f06: AnalogOut,          # 32 bit
    0x1f0c: AnalogOut,          # 64 bit

    0x2006: FlatIn,             # 32 bit
    0x200c: FlatIn,             # 64 bit
    0x2108: FlatIn,             # 32 bit
    0x2110: FlatIn,             # 64 bit
    0x220a: FlatIn,             # 32 bit
    0x2214: FlatIn,             # 64 bit
    0x230c: FlatIn,             # 32 bit
    0x2318: FlatIn,             # 64 bit
    0x240e: FlatIn,             # 32 bit
    0x241c: FlatIn,             # 64 bit
    0x2510: FlatIn,             # 32 bit
    0x2520: FlatIn,             # 64 bit
    0x2612: FlatIn,             # 32 bit
    0x2624: FlatIn,             # 64 bit
    0x2714: FlatIn,             # 32 bit
    0x2728: FlatIn,             # 64 bit
    0x2816: FlatIn,             # 32 bit
    0x282c: FlatIn,             # 64 bit
    0x2918: FlatIn,             # 32 bit
    0x2930: FlatIn,             # 64 bit
    0x2a1a: FlatIn,             # 32 bit
    0x2a34: FlatIn,             # 64 bit
    0x2b1c: FlatIn,             # 32 bit
    0x2b38: FlatIn,             # 64 bit
    0x2c1e: FlatIn,             # 32 bit
    0x2c3c: FlatIn,             # 64 bit
    0x2d20: FlatIn,             # 32 bit
    0x2d40: FlatIn,             # 64 bit
    0x2e22: FlatIn,             # 32 bit
    0x2e44: FlatIn,             # 64 bit
    0x2f24: FlatIn,             # 32 bit
    0x2f48: FlatIn,             # 64 bit

    0x3008: FlatOut,            # 32 bit
    0x3010: FlatOut,            # 64 bit
    0x310a: FlatOut,            # 32 bit
    0x3114: FlatOut,            # 64 bit
    0x320c: FlatOut,            # 32 bit
    0x3218: FlatOut,            # 64 bit
    0x330e: FlatOut,            # 32 bit
    0x331c: FlatOut,            # 64 bit
    0x3410: FlatOut,            # 32 bit
    0x3420: FlatOut,            # 64 bit
    0x3512: FlatOut,            # 32 bit
    0x3524: FlatOut,            # 64 bit
    0x3614: FlatOut,            # 32 bit
    0x3628: FlatOut,            # 64 bit
    0x3716: FlatOut,            # 32 bit
    0x372c: FlatOut,            # 64 bit
    0x3818: FlatOut,            # 32 bit
    0x3830: FlatOut,            # 64 bit
    0x391a: FlatOut,            # 32 bit
    0x3934: FlatOut,            # 64 bit
    0x3a1c: FlatOut,            # 32 bit
    0x3a38: FlatOut,            # 64 bit
    0x3b1e: FlatOut,            # 32 bit
    0x3b3c: FlatOut,            # 64 bit
    0x3c20: FlatOut,            # 32 bit
    0x3c40: FlatOut,            # 64 bit
    0x3d22: FlatOut,            # 32 bit
    0x3d44: FlatOut,            # 64 bit
    0x3e24: FlatOut,            # 32 bit
    0x3e48: FlatOut,            # 64 bit
    0x3f26: FlatOut,            # 32 bit
    0x3f4c: FlatOut,            # 64 bit

    0x4006: ParamIn,            # 32 bit
    0x400c: ParamIn,            # 64 bit
    0x5008: ParamOut,           # 32 bit
    0x5010: ParamOut,           # 64 bit

    0x4108: VectorIn,           # 32 bit
    0x4110: VectorIn,           # 64 bit
    0x420a: VectorIn,           # 32 bit
    0x4214: VectorIn,           # 64 bit
    0x430c: VectorIn,           # 32 bit
    0x4318: VectorIn,           # 64 bit
    0x440e: VectorIn,           # 32 bit
    0x441c: VectorIn,           # 64 bit
    0x4510: VectorIn,           # 32 bit
    0x4520: VectorIn,           # 64 bit
    0x4612: VectorIn,           # 32 bit
    0x4624: VectorIn,           # 64 bit
    0x4714: VectorIn,           # 32 bit
    0x4728: VectorIn,           # 64 bit
    0x4816: VectorIn,           # 32 bit
    0x482c: VectorIn,           # 64 bit
    0x4918: VectorIn,           # 32 bit
    0x4930: VectorIn,           # 64 bit
    0x4a1a: VectorIn,           # 32 bit
    0x4a34: VectorIn,           # 64 bit
    0x4b1c: VectorIn,           # 32 bit
    0x4b38: VectorIn,           # 64 bit
    0x4c1e: VectorIn,           # 32 bit
    0x4c3c: VectorIn,           # 64 bit
    0x4d20: VectorIn,           # 32 bit
    0x4d40: VectorIn,           # 64 bit
    0x4e22: VectorIn,           # 32 bit
    0x4e44: VectorIn,           # 64 bit
    0x4f24: VectorIn,           # 32 bit
    0x4f48: VectorIn,           # 64 bit

    0x510c: VectorOut,          # 32 bit
    0x5118: VectorOut,          # 64 bit
    0x5210: VectorOut,          # 32 bit
    0x5220: VectorOut,          # 64 bit
    0x5314: VectorOut,          # 32 bit
    0x5328: VectorOut,          # 64 bit
    0x5418: VectorOut,          # 32 bit
    0x5430: VectorOut,          # 64 bit
    0x551c: VectorOut,          # 32 bit
    0x5538: VectorOut,          # 64 bit
    0x5620: VectorOut,          # 32 bit
    0x5640: VectorOut,          # 64 bit
    0x5724: VectorOut,          # 32 bit
    0x5748: VectorOut,          # 64 bit
    0x5828: VectorOut,          # 32 bit
    0x5850: VectorOut,          # 64 bit
    0x592c: VectorOut,          # 32 bit
    0x5958: VectorOut,          # 64 bit
    0x5a30: VectorOut,          # 32 bit
    0x5a60: VectorOut,          # 64 bit
    0x5b34: VectorOut,          # 32 bit
    0x5b68: VectorOut,          # 64 bit
    0x5c38: VectorOut,          # 32 bit
    0x5c70: VectorOut,          # 64 bit
    0x5d3c: VectorOut,          # 32 bit
    0x5d78: VectorOut,          # 64 bit
    0x5e40: VectorOut,          # 32 bit
    0x5e80: VectorOut,          # 64 bit
    0x5f44: VectorOut,          # 32 bit
    0x5f88: VectorOut,          # 64 bit
}



#  -*- coding: utf-8 -*-
# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

# pylint: disable=invalid-name, unused-argument, no-value-for-parameter

from test.simulator.funcs import adr, memcpy, memset, sizeof
from test.simulator.runtime import Globals, Struct, Var, array, boolean, \
    byte, dword, program, real, string, word


class ST_DeviceInfo(Struct):
    TypCode = Var(word, 0)
    Size = Var(word, 0)
    Offset = Var(word, 0)
    Unit = Var(word, 0)
    Flags = Var(dword, 0)
    Params = Var(array(word, 1, 16), [0] * 16)
    Name = Var(string(34))
    Aux = Var(array(string(34), 0, 7))
    AbsMax = Var(real, 3.402823e38)
    AbsMin = Var(real, -3.402823e38)


class ST_Indexer(Struct):
    Request = Var(word, 0)
    Data = Var(array(word, 1, 17))


class Global(Globals):
    fMagic = Var(real, 2015.02, at='%MB0')
    iOffset = Var(word, 64, at='%MB4')

    stIndexer = Var(ST_Indexer, at='%MB64')

    Devices = Var(array(ST_DeviceInfo, 1, 3), [
        ST_DeviceInfo(TypCode=0x5008, Name='motor', Offset=100, Unit=0xfd04,
                      Params=[0, 0xc000, 0x000f, 0xf800, 0x001b, 0, 0, 0, 0x200]),
        ST_DeviceInfo(TypCode=0x1201, Name='rsvd_input'),
        ST_DeviceInfo(TypCode=0x1602, Name='rsvd_output', Offset=136),
    ])

    sPLCName = Var(string(34), 'jemsc')
    sPLCVersion = Var(string(34), '0.0.1alpha')
    iCycle = Var(word, 0)

g = Global()


@program(
    nDevices = Var(word),
    devnum = Var(word),
    infotype = Var(word),
    is_initialized = Var(boolean, False),
    itemp = Var(byte),
    tempofs = Var(word))
def Indexer(v):
    if not v.is_initialized:
        v.tempofs = g.iOffset + sizeof(g.stIndexer)
        v.nDevices = sizeof(g.Devices) // sizeof(g.Devices[1])
        v.itemp = 1
        while v.itemp <= v.nDevices:
            dev = g.Devices[v.itemp]
            # for i in range(8):
            dev.Flags[[0]] = len(dev.Aux[0]) > 0
            dev.Flags[[1]] = len(dev.Aux[1]) > 0
            dev.Flags[[2]] = len(dev.Aux[2]) > 0
            dev.Flags[[3]] = len(dev.Aux[3]) > 0
            dev.Flags[[4]] = len(dev.Aux[4]) > 0
            dev.Flags[[5]] = len(dev.Aux[5]) > 0
            dev.Flags[[6]] = len(dev.Aux[6]) > 0
            dev.Flags[[7]] = len(dev.Aux[7]) > 0
            if dev.Size < (dev.TypCode & 0xff) << 1:
                dev.Size = (dev.TypCode & 0xff) << 1
            if dev.Offset == 0:
                dev.Offset = v.tempofs
            else:
                v.tempofs = dev.Offset
            v.tempofs += dev.Size
            v.itemp += 1
        v.is_initialized = True

    if g.fMagic != 2015.02:
        g.fMagic = 2015.02
    if g.iOffset != 64:
        g.iOffset = 64

    v.devnum = g.stIndexer.Request & 0xff
    v.infotype = (g.stIndexer.Request >> 8) & 0x7f

    data = g.stIndexer.Data
    memset(adr(data), 0, sizeof(data))

    if v.devnum == 0:
        if v.infotype == 0:
            data[1] = 0
            data[2] = sizeof(g.stIndexer)
            data[3] = g.iOffset
            data[4] = 0
            data[5] = 0
            data[6] = 0
        elif v.infotype == 1:
            data[1] = sizeof(g.stIndexer)
        elif v.infotype == 4:
            memcpy(adr(data), adr(g.sPLCName),
                   min(sizeof(g.sPLCName), sizeof(data)))
        elif v.infotype == 5:
            memcpy(adr(data), adr(g.sPLCVersion),
                   min(sizeof(g.sPLCVersion), sizeof(data)))
    elif v.devnum <= v.nDevices:
        dev = g.Devices[v.devnum]
        if v.infotype == 0:
            data[1] = dev.TypCode
            data[2] = dev.Size
            data[3] = dev.Offset
            data[4] = dev.Unit
            data[5] = dev.Flags
            data[6] = dev.Flags >> 16
            memcpy(adr(data[7]), adr(dev.AbsMin), sizeof(dev.AbsMin))
            memcpy(adr(data[9]), adr(dev.AbsMax), sizeof(dev.AbsMax))
            memcpy(adr(data[11]), adr(dev.Name),
                   min(sizeof(dev.Name), sizeof(data) - 20))
        elif v.infotype == 1:
            data[1] = dev.Size
        elif v.infotype == 2:
            data[1] = dev.Offset
        elif v.infotype == 3:
            data[1] = dev.Unit
        elif v.infotype == 4:
            memcpy(adr(data), adr(dev.Name),
                   min(sizeof(dev.Name), sizeof(data)))
        elif v.infotype == 15:
            memcpy(adr(data), adr(dev.Params),
                   min(sizeof(dev.Params), sizeof(data)))
        elif v.infotype >= 0x10 and v.infotype <= 0x17:
            memcpy(adr(data), adr(dev.Aux[v.infotype - 0x10]),
                   min(sizeof(dev.Aux[v.infotype - 0x10]), sizeof(data)))

    if v.infotype == 127:
        data[1] = g.iCycle
    g.stIndexer.Request[[15]] = 1
    g.iCycle += 1

    #print(g.stIndexer.Request, data)


@program()
def Implementierung(v):
    pass


@program()
def Main(v):
    Indexer()
    Implementierung()



#  -*- coding: utf-8 -*-
# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************


class NumProxy:

    def __add__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__add__(other)

    def __sub__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__sub__(other)

    def __mul__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__mul__(other)

    def __floordiv__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__floordiv__(other)

    def __mod__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__mod__(other)

    def __and__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__and__(other)

    def __or__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__or__(other)

    def __xor__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__xor__(other)

    def __rshift__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__rshift__(other)

    def __lshift__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__lshift__(other)

    def __lt__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__lt__(other)

    def __le__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__le__(other)

    def __ge__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__ge__(other)

    def __gt__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__gt__(other)

    def __eq__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__eq__(other)

    def __ne__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value.__ne__(other)

    def __bool__(self):
        return self.value.__bool__()

    def __neg__(self):
        return self.value.__neg__()

    def __pos__(self):
        return self.value.__pos__()

    def __abs__(self):
        return self.value.__abs__()

    def __invert__(self):
        return self.value.__invert__()

    def __index__(self):
        return self.value.__index__()



#  -*- coding: utf-8 -*-
# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

import collections
import struct

from .util import NumProxy


class Memory:
    # Allocates addresses for variables.

    def __init__(self):
        # dyn is at 0x00000
        # %M* is at 0x10000
        # %I* is at 0x20000
        # %Q* is at 0x30000
        self.dyn_addr = 0
        self.allocations = []
        self.allocations_sorted = False

    def new(self, size):
        addr = self.dyn_addr
        self.dyn_addr += size
        return addr

    def map(self, addr, obj):
        """Map a new allocation."""
        self.allocations.append((obj.sizeof(), addr, obj))
        self.allocations_sorted = False

    def get(self, addr, size):
        """Return the object that addr belongs to, and the offset in it,
        if there are at least size bytes left.
        """
        if not self.allocations_sorted:
            self.allocations.sort()
            # print(self.allocations)
            self.allocations_sorted = True
        for (osize, oaddr, obj) in self.allocations:
            if oaddr <= addr and addr + size <= oaddr + osize:
                return (obj, addr - oaddr)
        raise RuntimeError('no value or addressing across value boundary')

    def read(self, addr, size):
        """Read memory."""
        (obj, offset) = self.get(addr, size)
        return obj.mem_read()[offset:offset+size]

    def write(self, addr, data):
        """Write memory."""
        (obj, offset) = self.get(addr, len(data))
        obj.mem_write(offset, data)


mem = Memory()


class Value:
    """Represents a place in memory for a variable."""

    @classmethod
    def alloc(cls, value, at=None):
        """Allocates an address for the value (if not given)."""
        if at is None:
            at = mem.new(cls.sizeof())
        val = cls(value, at)
        mem.map(at, val)
        return val

    @classmethod
    def default(cls, at=None):
        """Allocates a value with a default value."""
        return cls.alloc(cls.DEFAULT, at)

    @classmethod
    def unwrap(cls, value):
        """Retrieves the inner value from an rvalue assigned to this lvalue."""
        if isinstance(value, cls):
            return value.value
        return value

    @classmethod
    def sizeof(cls):
        """Returns the size of the value in memory."""
        raise NotImplementedError

    def __init__(self, value, addr):
        self.addr = addr
        self.assign(value)

    def __repr__(self):
        return repr(self.value)

    def assign(self, value):
        self.value = self.__class__.unwrap(value)

    def mem_read(self):
        raise NotImplementedError

    def mem_write(self, offset, data):
        raise NotImplementedError


class Integral(NumProxy, Value):
    DEFAULT = 0
    WIDTH = 0
    SIGNED = False
    MEMFMT = ''

    @classmethod
    def sizeof(cls):
        return cls.WIDTH // 8

    @classmethod
    def unwrap(cls, value):
        if isinstance(value, Integral):
            value = value.value
        limit = 1 << cls.WIDTH
        value %= limit
        if cls.SIGNED and value >= limit//2:
            value -= limit
        # if cls.SIGNED:
        #     limit = 1 << (cls.WIDTH - 1)
        #     if not -limit <= value < limit:
        #         raise RuntimeError('out of range assignment to %s: %s' % (
        #             cls.__name__, value))
        # else:
        #     if value < 0 or value >= 1 << cls.WIDTH:
        #         raise RuntimeError('out of range assignment to %s: %s' % (
        #             cls.__name__, value))
        return value

    def mem_read(self):
        return struct.pack(self.MEMFMT, self.value)

    def mem_write(self, offset, data):
        if offset != 0:
            raise RuntimeError('partial number write')
        self.value, = struct.unpack(self.MEMFMT, data)

    def __getitem__(self, i):
        # Bit access: a[[i]]
        return (self.value >> i[0]) & 1

    def __setitem__(self, i, val):
        mask = ~(1 << i[0])
        self.value = (self.value & mask) | ((val & 1) << i[0])


class byte(Integral):
    WIDTH = 8
    SIGNED = False
    MEMFMT = 'B'


class word(Integral):
    WIDTH = 16
    SIGNED = False
    MEMFMT = 'H'


class dword(Integral):
    WIDTH = 32
    SIGNED = False
    MEMFMT = 'I'


class lword(Integral):
    WIDTH = 64
    SIGNED = False
    MEMFMT = 'Q'


class boolean(Integral):
    WIDTH = 1
    SIGNED = False
    MEMFMT = 'B'

    @classmethod
    def sizeof(cls):
        return 1


class real(NumProxy, Value):
    DEFAULT = 0.0

    @classmethod
    def sizeof(cls):
        return 4

    def mem_read(self):
        return struct.pack('f', self.value)

    def mem_write(self, offset, data):
        if offset != 0:
            raise RuntimeError('partial number write')
        self.value, = struct.unpack('f', data)


class lreal(NumProxy, Value):
    DEFAULT = 0.0

    @classmethod
    def sizeof(cls):
        return 8

    def mem_read(self):
        return struct.pack('d', self.value)

    def mem_write(self, offset, data):
        if offset != 0:
            raise RuntimeError('partial number write')
        self.value, = struct.unpack('d', data)


class anystring(Value):
    SLEN = 0
    DEFAULT = ''

    @classmethod
    def sizeof(cls):
        return cls.SLEN

    @classmethod
    def unwrap(cls, value):
        if isinstance(value, anystring):
            value = value.value
        if len(value) > cls.SLEN:
            raise RuntimeError('string too long (%d chars max)' % cls.SLEN)
        return value

    def mem_read(self):
        return self.value.encode() + b'\0' * (self.SLEN - len(self.value))

    def mem_write(self, offset, data):
        if offset != 0:
            raise RuntimeError('partial string write')
        self.value = data.split('\0', 1)[0]

    def __len__(self):
        return self.value.__len__()


def string(slen):
    return type('string_%d' % slen, (anystring,), dict(SLEN=slen))


class anyarray(Value):
    LENGTH = 0
    IMIN = 0
    INNER = None
    DEFAULT = []

    @classmethod
    def sizeof(cls):
        return cls.LENGTH * cls.INNER.sizeof()

    def __init__(self, value, addr):
        self.addr = addr
        self.value = []
        value = self.__class__.unwrap(value)
        step = self.INNER.sizeof()
        if len(value) > self.LENGTH:
            raise RuntimeError('too many values in array assignment')
        for i in range(self.LENGTH):
            self.value.append(self.INNER.alloc(
                value[i] if i < len(value) else self.INNER.DEFAULT,
                at=addr + i*step))

    def assign(self, value):
        raise RuntimeError('array assign')

    def __getitem__(self, i):
        if not self.IMIN <= i < self.IMIN + self.LENGTH:
            raise RuntimeError('array access out of range')
        return self.value[i - self.IMIN]

    def __setitem__(self, i, val):
        if not self.IMIN <= i < self.IMIN + self.LENGTH:
            raise RuntimeError('array access out of range')
        self.value[i - self.IMIN].assign(val)

    def __len__(self):
        return self.LENGTH

    def __repr__(self):
        return '<%s>' % (', '.join('%r' % x for x in self.value))

    def mem_read(self):
        return b''.join(v.mem_read() for v in self.value)

    def mem_write(self, offset, data):
        step = self.INNER.sizeof()
        if offset % step != 0:
            raise RuntimeError('partial array element write')
        if len(data) % step != 0:
            raise RuntimeError('partial array element write')
        i = offset // step
        while data:
            d, data = data[:step], data[step:]
            self.value[i].mem_write(0, d)
            i += 1


def array(innertype, imin, imax):
    length = imax - imin + 1
    return type('array_%d' % length, (anyarray,), dict(LENGTH=length,
                                                       IMIN=imin,
                                                       INNER=innertype))


class Var:
    """Represents a field in a struct."""

    def __init__(self, dtype, default=None, *, at=None):
        self.dtype = dtype
        self.default = default if default is not None else dtype.DEFAULT
        self.at = None  # XXX
        if at is not None:
            if at.startswith('%MB'):
                self.at = int(at[3:]) + 0x10000
            elif at.startswith('%IB'):
                self.at = int(at[3:]) + 0x20000
            elif at.startswith('%QB'):
                self.at = int(at[3:]) + 0x30000
            else:
                raise RuntimeError('addr spec %s not supported' % at)

    def __get__(self, obj, obj_class):
        if obj is None:
            return self
        return obj.__dict__[self]

    def __set__(self, obj, value):
        if obj is None:
            return
        obj.__dict__[self].assign(self.dtype.unwrap(value))


class StructMeta(type):
    @classmethod
    def __prepare__(cls, _name, _bases):
        return collections.OrderedDict()

    def __init__(cls, _name, _bases, attrs):
        cls.VARS = []
        cls.OFFSET = {}
        size = 0
        for (varname, var) in attrs.items():
            if isinstance(var, Var):
                cls.VARS.append((varname, var))
                cls.OFFSET[varname] = size
                # XXX alignment!
                size += var.dtype.sizeof()
        cls.SIZE = size


class Struct(Value, metaclass=StructMeta):
    DEFAULT = Ellipsis

    @classmethod
    def sizeof(cls):
        return cls.SIZE

    def __init__(self, value=None, addr=None, **pvars):
        self.addr = addr
        for (name, var) in self.VARS:
            self.__dict__[var] = var.dtype.alloc(
                getattr(value, name, pvars.pop(name, var.default)),
                at=var.at or (addr + self.OFFSET[name] if addr is not None else None))
        if pvars:
            raise RuntimeError('unknown variable in struct: %s' % pvars)

    def assign(self, value):
        if value is Ellipsis:
            # XXX set defaults necessary?
            # keep defaults
            return
        if isinstance(value, self.__class__):
            # XXX reassign individual values!
            raise RuntimeError('struct reassign')
        raise RuntimeError('trying to assign struct of wrong kind')

    def __repr__(self):
        items = []
        for (name, var) in self.VARS:
            items.append((name, self.__dict__[var]))
        return '%s { %s }' % (self.__class__.__name__,
                              ', '.join('%s => %r' % x for x in items))

    def mem_read(self):
        return b''.join(self.__dict__[var].mem_read() for (_, var) in self.VARS)

    def mem_write(self, offset, data):
        skip = 0
        for (name, _) in self.VARS:
            if self.OFFSET[name] == offset:
                break
            skip += 1
        else:
            raise RuntimeError('partial struct element write')
        for (i, (name, var)) in enumerate(self.VARS):
            if i < skip:
                continue
            size = var.dtype.sizeof()
            if len(data) < size:
                raise RuntimeError('partial struct element write')
            d, data = data[:size], data[size:]
            self.__dict__[var].mem_write(0, d)
            if not data:
                break


class Globals(Struct):
    pass


def program(**pvars):
    def deco(func):
        var_struct = type('%s_vars' % func.__name__, (Struct,), pvars)
        instance = var_struct()

        def new_func():
            func(instance)

        new_func.is_program = True
        return new_func
    return deco



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Implements communication via ADS, the Beckhoff universal protocol."""

import re
import socket
import time
from struct import Struct, pack

from zapf import ApiError, CommError
from zapf.proto import Protocol

# Allow omitting the first 4 NetID bytes if it's the same as the IP address,
# but always require the port.
ADS_ADDR_RE = re.compile(r'ads://(.*?)/(\d+.\d+.\d+.\d+(.\d+.\d+)?)?:(\d+)$')

# TCP port to use for ADS.
ADS_PORT = 0xBF02

# AMS header including AMS/TCP header.
REQ_HEADER = Struct('<xxIQQHHIII')
# AMS reply header including error field.
REP_HEADER = Struct('<xxIQQHHIII')
# Reply to device info command.
DEVINFO = Struct('<BBH16s')

REQ_HEADER_SIZE = REQ_HEADER.size
REP_HEADER_SIZE = REP_HEADER.size
AMS_HEADER_SIZE = REQ_HEADER_SIZE - 6

# ADS commands.
ADS_DEVINFO = 1
ADS_READ    = 2
ADS_WRITE   = 3

# Formats for payload.
ADS_ADR_LEN = Struct('<III')
ADS_ERRID = Struct('<I')

# Index group for %M memory area.
INDEXGROUP_M = 0x4020

# UDP port
BECKHOFF_UDP_PORT = 0xBF03
# UDP message to set a route.
UDP_MESSAGE = Struct(
    '<I'     # magic
    '4x'     # pad
    'I'      # operation
    'Q'      # source netaddr
    'I'      # nitems=5
    'HH25s'  # desig=12 (routename), strlen=25, content
    'HH6s'   # desig=7 (netid), len=6, content
    'HH14s'  # desig=13 (username), len=14, content
    'HH2s'   # desig=2 (password), len=2, content
    'HH16s'  # desig=5 (host), len=16, content
)
# UDP magic header number.
UDP_MAGIC = 0x71146603
# UDP packet operations and data designators.
UDP_ADD_ROUTE = 6
UDP_PASSWORD = 2
UDP_HOST = 5
UDP_NETID = 7
UDP_ROUTENAME = 12
UDP_USERNAME = 13


def pack_net_id(netidstr, amsport):
    # convert a.b.c.d.e.f to a single integer
    amsnetid = i = 0
    for (i, x) in enumerate(netidstr.split('.')):
        amsnetid |= (int(x) << (8 * i))
    if i != 5:
        raise ValueError('incomplete NetID')

    # pack the whole address into a 64-bit integer
    return amsnetid | (amsport << 48)


class ADSProtocol(Protocol):
    OFFSETS = [0]

    def __init__(self, url, log):
        adr = ADS_ADDR_RE.match(url)
        if not adr:
            raise ApiError('invalid ADS address, must be '
                           'ads://host[:port]/amsnetid:amsport')
        host = adr.group(1)
        if ':' in host:
            host, port = host.split(':')
            port = int(port)
        else:
            port = ADS_PORT
        host_ip = socket.gethostbyname(host)

        if adr.group(3):    # full AMS netid specified
            netidstr = adr.group(2)
        elif adr.group(2):  # only first 4 specified
            netidstr = adr.group(2) + '.1.1'
        else:               # nothing specified, use IP
            netidstr = host_ip + '.1.1'
        amsport = int(adr.group(4))

        self._amsnetaddr = pack_net_id(netidstr, amsport)
        self._iphostport = (host, port)
        self._invoke_id = 1  # should be incremented for every request
        self._socket = None
        self._tried_route = False

        Protocol.__init__(self, url, log)

    def connect(self):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self._socket.connect(self._iphostport)
            self._myamsnetaddr = pack_net_id(
                self._socket.getsockname()[0] + '.1.1', 800)
            reply = self._comm(ADS_DEVINFO, b'', DEVINFO.size)
            v1, v2, v3, name = DEVINFO.unpack(reply)
            name = name.partition(b'\0')[0].decode('latin1')
            hw_name = '%s %d.%d.%d' % (name, v1, v2, v3)
            self.log.info('connected to %s', hw_name)
        except (OSError, ValueError) as err:
            self.log.exception('while connecting')
            raise CommError(f'could not connect to {self.url}: {err}') from err
        except CommError as err:
            # if we get a closed connection immediately, the route is not set
            # up correctly.  Try to fix that by setting a route via UDP.  If
            # the target TCP port is not default, we are not talking directly
            # to TwinCAT, so don't even try in that case.
            if str(err) == 'no data in read' and not self._tried_route and \
               self._iphostport[1] == ADS_PORT:
                self.log.warning('connection aborted, trying to set a route...')
                self._tried_route = True
                self._set_route(self._socket.getsockname())
                self.connect()
            raise
        self.connected = True

    def disconnect(self):
        try:
            self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()
        except OSError:
            pass
        self._socket = None
        self.connected = False

    def read(self, addr, length):
        if not self.connected:
            self.reconnect()
        payload = ADS_ADR_LEN.pack(INDEXGROUP_M, addr, length)
        try:
            return self._comm(ADS_READ, payload, 4 + length)[4:]
        except OSError as err:
            self.log.exception('during read')
            self.connected = False
            raise CommError('IO error during read: %s' % err) from err

    def write(self, addr, data):
        if not self.connected:
            self.reconnect()
        payload = ADS_ADR_LEN.pack(INDEXGROUP_M, addr, len(data)) + data
        try:
            self._comm(ADS_WRITE, payload, 0)
        except OSError as err:
            self.log.exception('during write')
            self.connected = False
            raise CommError('IO error during write: %s' % err) from err

    def _comm(self, cmd, payload, exp_len):
        """One ADS request-reply cycle."""
        invoke_id, self._invoke_id = self._invoke_id, self._invoke_id + 1
        msg = REQ_HEADER.pack(AMS_HEADER_SIZE + len(payload),
                              self._amsnetaddr, self._myamsnetaddr, cmd, 0x4,
                              len(payload), 0, invoke_id) + payload
        expected = REP_HEADER_SIZE + 4 + exp_len
        self._socket.sendall(msg)
        reply = b''
        rephdr = None
        while len(reply) < expected:
            data = self._socket.recv(expected - len(reply))
            if not data:
                raise CommError('no data in read')
            reply += data
            if len(reply) >= REP_HEADER_SIZE and not rephdr:
                rephdr = REP_HEADER.unpack_from(reply)
                if rephdr[4] != 0x5:
                    raise CommError('wrong flags in reply header')
                if rephdr[5] != len(data) - REP_HEADER_SIZE:
                    raise CommError('wrong length in reply header')
                if rephdr[6]:
                    raise CommError('error set in reply packet: %s' %
                                    self._translate_error(rephdr[6]))
                if rephdr[7] != invoke_id:
                    raise CommError('wrong InvokeID on reply packet')
        # check the error field that's only present if there is no general
        # error with the request (e.g. port not found)
        if reply[REP_HEADER_SIZE:REP_HEADER_SIZE+4] != b'\0\0\0\0':
            error = ADS_ERRID.unpack_from(reply, REP_HEADER_SIZE)[0]
            raise CommError('error set in reply packet: %s' %
                            self._translate_error(error))
        return reply[REP_HEADER_SIZE+4:]

    def _translate_error(self, errorcode):
        return ADS_ERRORS.get(errorcode,
                              'Unknown code %#04x' % errorcode)

    def _set_route(self, sockname):
        """Try to set up an ADS route on the target."""
        routename = 'zapf-' + sockname[0]
        udpsock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        mynetid = pack('<Q', self._myamsnetaddr)[:6]
        # The default password is different for different TwinCAT versions.
        # Just try both.
        for password in (b'', b'1'):
            message = UDP_MESSAGE.pack(
                UDP_MAGIC,
                UDP_ADD_ROUTE,
                self._myamsnetaddr,
                5,
                UDP_ROUTENAME, 25, routename.encode(),
                UDP_NETID, 6, mynetid,
                UDP_USERNAME, 14, b'Administrator',
                UDP_PASSWORD, 2, password,
                UDP_HOST, 16, sockname[0].encode(),
            )
            udpsock.sendto(message, (self._iphostport[0], BECKHOFF_UDP_PORT))
        time.sleep(0.5)


ADS_ERRORS = {
    0x001: 'Internal error',
    0x002: 'No Rtime',
    0x003: 'Allocation locked memory error',
    0x004: 'Insert mailbox error',
    0x005: 'Wrong receive HMSG',
    0x006: 'Target port not found',
    0x007: 'Target machine not found',
    0x008: 'Unknown command ID',
    0x009: 'Bad task ID',
    0x00A: 'No IO',
    0x00B: 'Unknown AMS command',
    0x00C: 'Win32 error',
    0x00D: 'Port not connected',
    0x00E: 'Invalid AMS length',
    0x00F: 'Invalid AMS NetID',
    0x010: 'Low installation level',
    0x011: 'No debug available',
    0x012: 'Port disabled',
    0x013: 'Port already connected',
    0x014: 'AMS Sync Win32 error',
    0x015: 'AMS Sync Timeout',
    0x016: 'AMS Sync AMS error',
    0x017: 'AMS Sync no index map',
    0x018: 'Invalid AMS port',
    0x019: 'No memory',
    0x01A: 'TCP send error',
    0x01B: 'Host unreachable',

    0x500: 'Router: no locked memory',
    0x502: 'Router: mailbox full',

    0x700: 'Error class: device error',
    0x701: 'Service is not supported by server',
    0x702: 'Invalid index group',
    0x703: 'Invalid index offset',
    0x704: 'Reading/writing not permitted',
    0x705: 'Parameter size not correct',
    0x706: 'Invalid parameter value(s)',
    0x707: 'Device is not in a ready state',
    0x708: 'Device is busy',
    0x709: 'Invalid context (must be in Windows)',
    0x70A: 'Out of memory',
    0x70B: 'Invalid parameter value(s)',
    0x70C: 'Not found (files, ...)',
    0x70D: 'Syntax error in command or file',
    0x70E: 'Objects do not match',
    0x70F: 'Object already exists',
    0x710: 'Symbol not found',
    0x711: 'Symbol version invalid',
    0x712: 'Server is in invalid state',
    0x713: 'AdsTransMode not supported',
    0x714: 'Notification handle is invalid',
    0x715: 'Notification client not registered',
    0x716: 'No more notification handles',
    0x717: 'Size for watch too big',
    0x718: 'Device not initialized',
    0x719: 'Device has a timeout',
    0x71A: 'Query interface failed',
    0x71B: 'Wrong interface required',
    0x71C: 'Class ID is invalid',
    0x71D: 'Object ID is invalid',
    0x71E: 'Request is pending',
    0x71F: 'Request is aborted',
    0x720: 'Signal warning',
    0x721: 'Invalid array index',

    0x740: 'Error class: client error',
    0x741: 'Invalid parameter at service',
    0x742: 'Polling list is empty',
    0x743: 'Var connection already in use',
    0x744: 'Invoke ID in use',
    0x745: 'Timeout elapsed',
    0x746: 'Error in Win32 subsystem',
    0x748: 'ADS port not opened',
    0x750: 'Internal error in ADS sync',
    0x751: 'Hash table overflow',
    0x752: 'Key not found in hash',
    0x753: 'No more symbols in cache',
}



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Contains all necessary constants and mappings from the spec."""

from zapf.util import BitFields, CommandMap, ParamMap, UncasedMap

SUPPORTED_MAGICS = ['2015_02']

OFFSET_ADDR = 4

INDEXER_DEV = 0

INFO_STRUCT  = 0
INFO_SIZE    = 1
INFO_ADDR    = 2
INFO_UNIT    = 3
INFO_NAME    = 4
INFO_VERSION = 5
INFO_AUTHOR1 = 6
INFO_AUTHOR2 = 7
INFO_PARAMS  = 15
INFO_AUX1    = 16
INFO_CYCLE   = 127


UNIT_CODES    = ('', 'V', 'A', 'W', 'm', 'g', 'Hz', 'T', 'K', 'degC', 'degF',
                 'bar', 'deg', 'Ohm', 'm/s', 'm^2/s', 'm^3/s', 's', 'cts',
                 'bar/s', 'bar/s^2', 'F', 'H', 'l/min')
UNIT_EXPONENT = {0: '', 2: 'h', 3: 'k', 6: 'M', 9: 'G', 12: 'T', 15: 'P',
                 18: 'E', -1: 'd', -2: 'c', -3: 'm', -6: 'u', -9: 'n',
                 -12: 'f', -15: 'a'}
UNIT_SPECIAL  = {(-2, 0): '%', (-3, 16): 'l/s'}

FLOAT32_MAX = 3.402823e+38
FLOAT64_MAX = 1.7976931348623158e+308

StatusStruct = BitFields(STATE=(15, 12), REASON=(11, 8), AUX=(7, 0), ISERR=15)
ParamControl = BitFields(CMD=(15, 13), SUBINDEX=(12, 8), IDX=(7, 0))

# The first 8 AUX bits
ReasonMap = [
    '',
    'inhibit',
    'timeout',
    '<Reason 3>',
    'lower limit reached',
    '<Reason 5>',
    '<Reason 6>',
    '<Reason 7>',
    'higher limit reached',
    '<Reason 9>',
    '<Reason 10>',
    '<Reason 11>',
    'both limits reached',
    '<Reason 13>',
    '<Reason 14>',
    '<Reason 15>',
]

ParamCMDs = UncasedMap(
    INIT=0,
    DO_READ=1,
    DO_WRITE=2,
    BUSY=3,
    DONE=4,
    ERR_NO_IDX=5,
    ERR_RO=6,
    ERR_RETRY=7,
)

PLCStatus = UncasedMap(
    RESET=0,
    IDLE=1,
    DISABLED=2,
    WARN=3,
    START=5,
    BUSY=6,
    STOP=7,
    ERROR=8,
    DIAGNOSTIC_ERROR=13,
)

# The first (numbered) parameter whose value is float instead of integer.
FIRST_FLOAT_PARAM = 30

# Parameter 0 should not be used.
Parameters = ParamMap(
    UNUSED=0,
    Mode=1,
    Microsteps=2,
    CoderBits=3,
    ExtStatus=4,
    # HwOffset=5,
    AbsMin=30,  # deprecated
    AbsMax=31,  # deprecated
    UserMin=32,
    UserMax=33,
    WarnMin=34,
    WarnMax=35,
    TimeoutTime=36,
    MaxTravelDist=37,
    AccelTime=38,
    Offset=40,
    BlockSize=43,
    Opening=44,
    PidP=51,
    PidI=52,
    PidD=53,
    DragError=55,
    Hysteresis=56,
    Holdback=57,
    RefSpeed=58,
    Jerk=59,
    Speed=60,
    Accel=61,
    IdleCurrent=62,
    RampCurrent=63,
    MoveCurrent=64,
    StopCurrent=65,
    # AbortDecel=66,
    LegacyMicrosteps=67,  # deprecated
    Slope=68,
    RefPos=69,
    Setpoint=70,
)

SpecialFunctions = CommandMap(
    FactoryReset=128,
    RefNeg=131,  # deprecated
    RefPos=132,  # deprecated
    Ref=133,
    SetPosition=137,
    ContMove=142,
)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

import logging
import struct
import threading
import time

import pytest

from zapf import CommError
from zapf.io import PlcIO
from zapf.proto import Protocol

from test.simulator import runtime, testplc


def prettify(data):
    if len(data) == 2:
        val = struct.unpack('H', data)[0]
        return f'{val} {val:#x}'
    elif len(data) == 4:
        val_i = struct.unpack('I', data)[0]
        val_f = struct.unpack('f', data)[0]
        return f'{val_i} {val_i:#x} {val_f}'
    elif len(data) == 8:
        val_i = struct.unpack('Q', data)[0]
        val_f = struct.unpack('d', data)[0]
        return f'{val_i} {val_i:#x} {val_f}'
    return ''


class TestProtocol(Protocol):
    OFFSETS = [0, 0x6000, 0x8000, 0x10000]

    def __init__(self, url, log, read, write):
        Protocol.__init__(self, url, log)
        self._handle_read = read
        self._handle_write = write

    def connect(self):
        self.connected = True

    def disconnect(self):
        self.connected = False

    def read(self, addr, length):
        try:
            data = self._handle_read(self.offset + addr, length)
        except Exception as err:
            self.log.warning(f'R {addr:#x} {length} !! {err}')
            raise CommError(f'read failed: {err}') from err
        else:
            self.log.info(f'R {addr:#x} {length} -> {data} {prettify(data)}')
            return data

    def write(self, addr, data):
        try:
            self._handle_write(self.offset + addr, data)
        except Exception as err:
            self.log.warning(f'W {addr:#x} {data} {prettify(data)} !! {err}')
            raise CommError(f'write failed: {err}') from err
        else:
            self.log.info(f'W {addr:#x} {data} {prettify(data)} ok')


@pytest.fixture(scope='function', autouse=True)
def plc_io(caplog):
    caplog.set_level(logging.INFO)

    cond = threading.Condition()
    stopflag = False

    def thread():
        while not stopflag:
            with cond:
                testplc.Main()  # pylint: disable=no-value-for-parameter
                cond.notify()
            time.sleep(.005)

    def handle_read(addr, length):
        with cond:
            cond.wait()
            return runtime.mem.read(addr, length)

    def handle_write(addr, data):
        with cond:
            cond.wait()
            runtime.mem.write(addr, data)

    plc = threading.Thread(target=thread, daemon=True)
    plc.start()

    plogger = logging.getLogger('simplc')
    proto = TestProtocol('test://', plogger, handle_read, handle_write)
    yield PlcIO(proto, logging.root)

    stopflag = True
    plc.join()



import logging

import zapf.io
import zapf.scan

logging.basicConfig(level=logging.DEBUG)

#URL = 'modbus://127.0.0.1:5002/0'
#URL = 'ads://ccr12.ictrl.frm2/5.53.35.202:800'
URL = 'tango://ccr12.ictrl.frm2:10000/box/plc/modbus'

scan = zapf.scan.Scanner(URL, logging.root)
for name, addr, typecode, info in scan.scan_devices():
    pass
